/************************************************************************/
/* zen2time.c                                                           */
/*                                                                      */
/* Calculate time for given solar zenith angle.                         */
/*                                                                      */
/* Author: Bernhard Mayer,                                              */
/*        Fraunhofer Institute for Atmospheric Environmental Research,  */
/*        82467 Garmisch-Partenkirchen,                                 */
/*        Germany                                                       */
/*                                                                      */
/*----------------------------------------------------------------------*/
/* Copyright (C) 1995 Bernhard Mayer                                    */
/*                                                                      */
/* This program is free software; you can redistribute it and/or modify */
/* it under the terms of the GNU General Public License as published by */
/* the Free Software Foundation; either version 1, or (at your option)  */
/* any later version.                                                   */
/*                                                                      */
/* This program is distributed in the hope that it will be useful,      */
/* but WITHOUT ANY WARRANTY; without even the implied warranty of       */
/* MERCHANTABILITY of FITNESS FOR A PARTICULAR PURPOSE. See the         */
/* GNU General Public License for more details.                         */
/*                                                                      */
/* To obtain a copy of the GNU General Public License write to the      */
/* Free Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139,   */
/* USA.                                                                 */
/*----------------------------------------------------------------------*/
/************************************************************************/

/*    Changes by Arve Kylling, 25 Jul 1997                              */
/*     i)  Added -p option  to turn of printing of messages.            */ 
/*     ii) Removed line containing optmode, not needed with the getopt  */
/*         I have.                                                      */

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <getopt.h>

#include "sun.h"


#define IFU_LATITUDE   47.48
#define IFU_LONGITUDE -11.07
#define IFU_LONG_STD  -15.00



/*************************************************************/
/* print usage information                                   */
/*************************************************************/

static void print_usage (char *filename)
{
  fprintf (stderr, "ZEN2TIME 1.99b - calculate standard time from solar zenith angle\n\n");
  fprintf (stderr, "written by  Bernhard Mayer,\n");
  fprintf (stderr, "            Fraunhofer-Institute for Atmospheric Research (IFU)\n");
  fprintf (stderr, "            82467 Garmisch-Partenkirchen, Germany\n");
  fprintf (stderr, "Version 1.99b finished December 20, 1995\n\n");
  fprintf (stderr, "Be aware that this program is a beta version! Please report any\n");
  fprintf (stderr, "kind of error to the author, including the error message and the\n");
  fprintf (stderr, "database being processed when the error occured. Thanks!\n\n");
  fprintf (stderr, "USAGE: %s [options] <day> <month> <zenith>\n\n", filename);
  fprintf (stderr, "Optional arguments:\n");
  fprintf (stderr, "  -h               display help message\n");
  fprintf (stderr, "  -a               Latitude  (North positive)\n");
  fprintf (stderr, "  -o               Longitude (West positive)\n");
  fprintf (stderr, "  -s               Standard Longitude\n\n");
  fprintf (stderr, "If the coordinates are omitted, the location of IFU is used for\n");
  fprintf (stderr, "the calculation:\n");
  fprintf (stderr, "  Latitude 47.48, Longitude -11.07, Standard Longitude.\n");
}



/*************************************************************/
/* parse command line options                                */
/*************************************************************/

static int get_options (int argc, char **argv,
			char *programname,
			int *doy,
			int *print,
			double *zenith,
			double *latitude,
			double *longitude,
			double *long_std)
{
  int c=0;
  char *dummy=NULL;

  int day   = 0;
  int month = 0;

  char lat=0, lon=0, std=0;


  /* reset parameters to default values */
  *latitude  = IFU_LATITUDE;
  *longitude = IFU_LONGITUDE;
  *long_std  = IFU_LONG_STD;
  

  /* save name of program */
  strncpy (programname, argv[0], FILENAME_MAX);
  
  /* get command line arguments */
  /* The following line is not needed with the getopt.{c,h} I have */
  /* Arve Kylling, Jul 25, 1997 */
  /* optmode = GETOPT_ANY; */

  while ((c=getopt (argc, argv, "ha:o:ps:")) != EOF)  {
    switch(c)  {
    case 'h':  /* help */
      print_usage (programname);
      return (-1);
      break;
    case 'a':  /* latitude  */  
      *latitude  = (double) strtod (optarg, &dummy);
      lat=1;
      break;   
    case 'o':  /* longitude */  
      *longitude = (double) strtod (optarg, &dummy);
      lon=1;
      break;   
    case 'p':  /* no print messages */  
      *print = 0;
      break;   
    case 's':  /* standard longitude */  
      *long_std = (double) strtod (optarg, &dummy);
      std=1;
      break;   
    case '?':
      print_usage (programname);
      return (-1);
      break;
    default:
      print_usage (programname);
      return (-1);
    }
  }
      

  /* check number of remaining command line arguments */
  if (argc - optind != 3)  {
    print_usage (programname);
    return -1;
  }
  
  /* check arguments */
  if (lat&!lon || lat&!std || lon&!lat || lon&!std || std&!lat || std&!lon) {
    print_usage (programname);
    fprintf (stderr, "\n** Specify all three parameters, latitude, longitude and **\n");
    fprintf (stderr,   "** standard longitude!!                                  **\n");
    
    *latitude  = IFU_LATITUDE;
    *longitude = IFU_LONGITUDE;
    *long_std  = IFU_LONG_STD;

    return -1;
  }


  /* check coordinates range */
  if (*latitude<-90 || *latitude>90)  {
    fprintf (stderr, "\n** Latitude out of range **\n");
    return -1;
  }

  if (*longitude<-180 || *longitude>180)  {
    fprintf (stderr, "\n** Longitude out of range **\n");
    return -1;
  }

  if (*long_std<-180 || *long_std>180)  {
    fprintf (stderr, "\n** Standard Longitude out of range **\n");
    return -1;
  }


  /* get command line arguments */
  day    = strtod (argv[optind+0], &dummy);
  month  = strtod (argv[optind+1], &dummy);
  *zenith = strtod (argv[optind+2], &dummy);

  if (*zenith<0 || *zenith>180)  {
    fprintf (stderr, "\n** Solar zenith angle out of range **\n");
    return -1;
  }

  /* calculate day of year */
  if ((*doy = day_of_year(day, month)) < 0)  {
    print_usage (programname);
    fprintf (stderr, "\n** Date out of range **\n");
    return -1;
  }


  return 0;  /* if o.k. */
}






int main(int argc, char ** argv)
{
  char programname[FILENAME_MAX]="";

  double latitude  = 0;
  double longitude = 0;
  double long_std  = 0;
  double zenith    = 0;

  double stunde1=0, stunde2=0;

  int doy   = 0;
  int hour  = 0;
  int min   = 0;
  int sec   = 0;

  int time1=0, time2=0;

  int status=0;
  int print =1;

  char dummy[9] = "";

  
  /* get command line options */
  status = get_options (argc, argv,
			programname,
			&doy, &print, &zenith,
			&latitude, &longitude, &long_std);

  if (status!=0)
    return status;





  /* calculate time */
  if (zenith2time (doy, zenith, 
		   latitude, longitude, long_std, 
		   &time1, &time2) != 0)  {
    fprintf (stderr, "** Sorry, zenith angle too low for this day! **\n");
    fprintf (stdout, "  %g  %g  %g\n", zenith, NAN, NAN);
    return (-1);
  }


  if ( print ) {
    fprintf (stderr, "Coordinates: Lat. %8.4f, Long. %8.4f, Standard Long. %8.4f\n",
	     latitude, longitude, long_std);
    fprintf (stderr, "  Solar Zenith  Angle:   %f\n\n", zenith);
  }

  stunde1 = (double) time1 / 3600.0;
  stunde2 = (double) time2 / 3600.0;

  hour   = time1 / 3600;
  time1 %= 3600;
  min    = time1 / 60;
  time1 %= 60;
  sec    = time1;

  if ( print ) {
    fprintf (stderr, "  Standard time #1:      %s MEZ\n", time2str(dummy, hour, min, sec));
  }

  hour   = time2 / 3600;
  time2 %= 3600;
  min    = time2 / 60;
  time2 %= 60;
  sec    = time2;

  if ( print ) {
    fprintf (stderr, "  Standard time #2:      %s MEZ\n", time2str(dummy, hour, min, sec));

    fprintf (stderr, "\n  Printing to stdout: zenith angle, morning time, afternoon time\n");
  }
  fprintf (stdout, "  %g  %g  %g\n", 
	   zenith, stunde1, stunde2); 

  return 0;
}
